# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
import avro.protocol
import avro.ipc
import copy
import logging
import os
import socket
import sys
import time
from cmf.throttling_logger import ThrottlingLogger
from httplib import BadStatusLine
from urlparse import urlparse

LOG = logging.getLogger( __name__ )
THROTTLED_LOG = ThrottlingLogger(LOG)
NAVIGATOR_AGENT_PROTOCOL = avro.protocol.parse(
  file(os.path.join(os.path.dirname(__file__), "navigator.avpr")).read())

class ImpaladAuditSender(object):
  """
  This class parses a line from impalad audit log and then creates one or
  more audit events from it. The events are then sent to the audit server.
  """
  def __init__(self, audit_server_url, service_name):
    parsed_url = urlparse(audit_server_url)
    netloc = parsed_url[1]
    self._host_name, self._port = netloc.split(':', 1)
    self._service_name = service_name
    self._transceiver = None
    self._requestor = None

  def close(self):
    if self._transceiver is not None:
      try:
        self._transceiver.close()
      except:
        pass # ignore

      self._requestor = None
      self._transceiver = None
      
  def send_audit_events(self, audit_event_batch):
    """
    Method to parse the audit log line and send one or multiple audit events
    to the audit server.
    Returns True if the audit was successfully sent to the audit server, False
    otherwise.
    """
    assert len(audit_event_batch) > 0
    
    ret = True
    try:
      if self._transceiver is None:
        self._transceiver = avro.ipc.HTTPTransceiver(self._host_name, self._port)
        self._requestor = avro.ipc.Requestor(NAVIGATOR_AGENT_PROTOCOL, self._transceiver)
        LOG.info("Successfully connected to audit server at url: %s, port: %s" %
                (self._host_name, self._port))
    except socket.error, ex:
      # Unable to connect to audit server. wait for it to come up
      THROTTLED_LOG.info("Unable to communicate with audit server: %s" % (ex, ))
      return False
      
    try:
      events = []
      for (_, _, audit_event) in audit_event_batch:
        for avro_audit_event in parse_impalad_audit_event(self._service_name, audit_event):
          events.append(dict(event=avro_audit_event))
      
      LOG.debug("Sending %d audit events to Audit server." % (len(events), ))
      wae = {'serviceName' : self._service_name,
             'events' : events,
             'counters' : None
             }
      response = self._requestor.request('writeAuditEvents', dict(request=wae))
      if response['success']:
        LOG.debug("Successfully sent %d events to audit server" % 
                  (len(events), ))
      else:
        THROTTLED_LOG.error("Error occurred when sending audit to audit server")
        ret = False
    except BadStatusLine, ex:
      # this means we lost connection to nav server and need to restablish it
      # Return False and let caller retry
      self.close()
      ret = False
    except Exception, ex:
      THROTTLED_LOG.exception("Unable to send message to server.")
      self.close()
      ret = False
      
    return ret
  
def parse_impalad_audit_event(service_name, audit_log_json):
  """
  Method to parse a JSON audit log and create one or multiple audit events
  from it. 
  """
  # audit_log_json is a dict with only one entry:
  #    key is timestamp of event, value is the audit event
  timestamp = audit_log_json.keys()[0]
  audit_json = audit_log_json[timestamp]
  
  # audit event start time is like 2013-08-05 14:10:05.324341000. Convert it
  # to seconds since epoch (python 2.4 limitation, we need to drop fractional
  # second part)
  start_time = audit_json['start_time'].strip().split(".")[0]
  start_time_struct = time.strptime(start_time, "%Y-%m-%d %H:%M:%S")
  
  audit_event = dict()
  audit_event['serviceName'] = service_name
  audit_event['username'] = audit_json['user']
  audit_event['impersonator'] = audit_json['impersonator']
  audit_event['ipAddress'], port = get_ipaddress_port(audit_json['network_address'])
  audit_event['operation'] = audit_json['statement_type']
  # server needs time in milliseconds
  audit_event['eventTime'] = int(time.mktime(start_time_struct) * 1000)
  audit_event['operationText'] = audit_json['sql_statement']
  audit_event['allowed'] = not audit_json['authorization_failure']
  audit_event['queryId'] = audit_json['query_id']
  audit_event['sessionId'] = audit_json['session_id']
  audit_event['status'] = audit_json['status']

  if audit_json['catalog_objects']:
    for catalog_object in audit_json['catalog_objects']:
      object_type = catalog_object['object_type']
      db_name, table_name = (None, catalog_object['name'])
      if object_type in ["TABLE", "VIEW"]:
        db_name, table_name = table_name.split(".")
      
      audit_event['databaseName'] = db_name
      audit_event['tableName'] = table_name
      audit_event['objectType'] = object_type
      audit_event['privilege'] = catalog_object['privilege']
      yield copy.copy(audit_event)
  else:
    yield copy.copy(audit_event)

def get_ipaddress_port(network_address):
  """
  Method to parse IP Address provided in Impala log and get IP Address
  and port from it.
  """
  ip_address = None
  port = None
  if network_address is not None:
    ip_address, port = network_address.rsplit(":", 1)

  return ip_address, port
