# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

try:
  import json
except ImportError:
  import simplejson as json

import logging
import os
import threading
import time

from cmf.monitor.count_down_latch import CountDownLatch
from cmf.monitor.datanode import DataNodeMonitor
from cmf.monitor.firehoses import Firehoses
from cmf.monitor.host import HostMonitor
from cmf.monitor.namenode import NameNodeMonitor
from cmf.monitor.secondarynamenode import SecondaryNameNodeMonitor
from cmf.monitor.jobtracker import JobTrackerMonitor
from cmf.monitor.tasktracker import TaskTrackerMonitor, init_tasktracker_monitor
from cmf.monitor.master import MasterMonitor
from cmf.monitor.regionserver import RegionServerMonitor
from cmf.monitor.servicemonitor import ServiceMonitorMonitor
from cmf.monitor.activitymonitor import ActivityMonitorMonitor
from cmf.monitor.eventserver import EventServerMonitor
from cmf.monitor.alertpublisher import AlertPublisherMonitor
from cmf.monitor.reportsmanager import ReportsManagerMonitor
from cmf.monitor.navigator import NavigatorMonitor
from cmf.monitor.hostmonitor import HostMonitorMonitor
from cmf.monitor.journalnode import JournalNodeMonitor
from cmf.monitor.zookeeper import ZooKeeperServerMonitor
from cmf.monitor.flume import FlumeAgentMonitor
from cmf.monitor.impalad import ImpaladMonitor
from cmf.monitor.statestore import StateStoreMonitor
from cmf.monitor.hbasethriftserver import HbaseThriftServerMonitor
from cmf.monitor.hbaserestserver import HbaseRestServerMonitor
from cmf.monitor.httpfs import HttpFSMonitor
from cmf.monitor.failovercontroller import FailoverControllerMonitor
from cmf.monitor.host_update import HostUpdate
from cmf.monitor.registry import MonitorRegistry
from cmf.monitor.wakeable_thread import WakeableThread
from cmf.monitor.host.filesystem_map import FilesystemMap
from cmf.monitor.json_metrics_utils import build_metric_accessors

LOG = logging.getLogger("monitor")

_METRIC_SCHEMA = 'schema.json'
_METRIC_SCHEMA_PATH = os.path.join(os.path.dirname(__file__), _METRIC_SCHEMA)

class MonitorDaemon(object):
  """
  The MonitorDaemon periodically reports metrics to the firehoses.
  """
  def __init__(self, hostname, host_id, filesystem_map=None, agent=None):
    self._agent = agent
    self._hostname = hostname
    self._host_id = host_id
    self._registry = MonitorRegistry()

    # This is set by the agent while it processes heartbeats, so folks should
    # take care while accessing it.
    self.enabled_metric_reporters = []

    # This thread is responsible for sending reports to the various firehoses.
    # It is awoken by the scheduling thread on a regular schedule (to send full
    # reports) and at special times when new monitors have been created and are
    # ready to send their first report (partial reports).
    self._reporting_thread = \
        WakeableThread("MonitorDaemon-Reporter", self._report)
    # This thread is responsible for waking up the reporting thread when it is
    # appropriate. It also drives detecting when new monitors are ready to send
    # their first report.
    self._scheduling_thread = \
        WakeableThread("MonitorDaemon-Scheduler", self._schedule)

    # This is the time at which we last woke the reporter thread to have it send
    # a full report.
    self._last_full_report_scheduled = None
    # This flag is set by the scheduling thread to indicate that the reporting
    # thread should send a full report when it next runs. It is cleared by the
    # reporting thread.
    self._full_report = False

    # This lock is used to protect access to the various new monitor members
    self._new_monitors_lock = threading.Lock()
    # This is a map from monitor object to 'expiration time'. A monitor is put
    # into this map when it is first created and it stays in the map until it
    # has either indicated that it is ready to send reports or the expiration
    # time has passed.
    self._new_monitors = {}
    # When a new monitor is ready to send a report, it is moved from the map
    # above to this list.
    self._new_monitors_ready = []

    # This object manages our connections to and is used to communicate with the
    # various firehoses.
    self.firehoses = Firehoses(self._host_id, self._hostname)

    if filesystem_map is None:
      self._fs_map = FilesystemMap()
    else:
      self._fs_map = filesystem_map

    init_tasktracker_monitor()
    self._host_monitor = HostMonitor(self, self._registry)

    if agent is None:
      # The agent is null during unit test runs, the timeouts don't matter in
      # that case, so just use whatever.
      self._metrics_url_timeout_seconds = 30.0
      self._task_metrics_timeout_seconds = 10.0
    else:
      self._metrics_url_timeout_seconds = agent.metrics_url_timeout_seconds
      self._task_metrics_timeout_seconds = agent.task_metrics_timeout_seconds

    LOG.info("Importing metric schema from file %s" % _METRIC_SCHEMA_PATH)
    self.metric_accessors = \
        build_metric_accessors(json.load(file(_METRIC_SCHEMA_PATH)))

  def get_agent(self):
    return self._agent

  def start(self):
    self._host_monitor.ensure_collecting()
    self._reporting_thread.start()
    # We force the scheduling thread to run immediately.
    self._scheduling_thread.wake(CountDownLatch(1))
    self._scheduling_thread.start()

  def should_stop(self):
    self._host_monitor.stop()
    self._reporting_thread.should_stop()
    self._scheduling_thread.should_stop()

  def get_reporting_interval_seconds(self):
    return self.firehoses.get_reporting_interval_seconds()

  def get_filesystem_map(self):
    return self._fs_map

  def record_heartbeat_timestamps(self,
                                  ts_agent_send,
                                  ts_agent_recv,
                                  ts_master_send,
                                  ts_master_recv):

    self._host_monitor.record_heartbeat_timestamps(
        ts_agent_send, ts_agent_recv,
        ts_master_send, ts_master_recv)

  def update_host_config(self, host_config):
    """
    Called by agent after a heartbeat to update host config passed from SCM.
    """
    self._host_monitor.update_host_config(host_config)

  def update_firehose_connection_info(self, firehose_connection_info):
    """Updates connection information to the various firehoses."""
    self.firehoses.update_firehose_connection_info(firehose_connection_info)

  def send_impala_query_update(self, impala_query_update):
    """Sends an impala query update to the host monitor."""
    self.firehoses.send_smon_update(
      service_updates=[],
      role_updates=[],
      impala_query_updates=[impala_query_update.to_avro_dict()])

  def _report(self):
    # Whether we are doing a full report or a partial report, we can clear the
    # list of new montitors that are ready.
    self._new_monitors_lock.acquire()
    new_monitors_ready = self._new_monitors_ready
    self._new_monitors_ready = []
    self._new_monitors_lock.release()

    if self._full_report:
      self._full_report = False
      monitors = self._registry.get_registry()
    else:
      monitors = new_monitors_ready

    # Filter out the monitors which have metric reporting disabled. Reporting
    # might be disabled for some monitors if say the customer lacks a particular
    # license.
    enabled = self.enabled_metric_reporters
    monitors = filter(lambda x: x.get_subject_type() in enabled, monitors)

    if len(monitors) > 0:
      self._report_for_monitors(monitors)

  def _report_for_monitors(self, participants):
    """Collect reports from the Monitors and send them to the firehoses"""

    # Update the process information of each of our monitors if necessary. Note
    # that if we are using supervisord events, then there's no need to do this
    # as any change in process status will have been communicated via an event
    # and the monitors will be updated by that event handling.
    if not self._agent.monitoring_supervisord_events():
      self._agent.update_monitor_process_information()

    # We signal each monitor in turn to start collecting then wait for each to
    # signal its reports are ready. We will wait at most the max_collection_wait
    # of the agent.
    latch = CountDownLatch(len(participants))
    for monitor in participants:
      monitor.start_collecting(latch)
    latch.await(self._agent.max_collection_wait_seconds)

    host_update = None
    role_updates = [ ]
    attempt_updates = [ ]
    service_updates = [ ]

    for monitor in participants:
      # We build a single large list of attempt updates. Each monitor can add to
      # it, but we store their updates in a single flat list.
      attempts = monitor.get_attempt_updates()
      if len(attempts) > 0:
        attempt_updates.extend(attempts)
        LOG.debug("%s reports attempt metrics: %s", monitor, attempts)

      # Each monitor gets to send a role update.
      role_update = self._safe_get_role_update(monitor)
      if role_update is not None:
        LOG.debug("%s reports role metrics: %s", monitor, role_update)
        role_updates.append(role_update.to_avro_dict())

      # Each monitor gets to send a service update.
      service_update = self._safe_get_service_update(monitor)
      if service_update is not None:
        LOG.debug("%s reports service metrics: %s", monitor, service_update)
        service_updates.append(service_update.to_avro_dict())

      if isinstance(monitor, HostMonitor):
        update = self._safe_get_host_update(monitor)
        LOG.debug("%s reports host metrics: %s", monitor, update)
        host_update = update.to_avro_dict()

    self.firehoses.send_amon_update(attempt_updates)
    self.firehoses.send_smon_update(service_updates, role_updates, None)
    self.firehoses.send_hmon_update(host_update)

  # Wrapper call guaranteed not to throw.
  def _safe_get_role_update(self, monitor):
    try:
      return monitor.get_role_update()
    except:
      LOG.exception("Could not get role update from: %s" % (monitor,))
      return None

  # Wrapper call guaranteed not to throw.
  def _safe_get_service_update(self, monitor):
    try:
      return monitor.get_service_update()
    except:
      LOG.exception("Could not get service update from: %s" % (monitor,))
      return None

  # Wrapper call guaranteed not to throw.
  def _safe_get_host_update(self, monitor):
    try:
      return monitor.get_host_update()
    except:
      LOG.exception("Could not get host update from: %s" % (monitor,))
      return None

  def make_monitor(self, process_name, full_process_name, cgroups_root):
    """
    make_monitor(process_name, full_process_name, has_cgroups, cgroup_root) -> MonitorInterface

    The returned monitor is not running yet. The caller needs to
    refresh() it to activate it.

    Returns None if monitoring is not supported.
    """
    cgroup_name = full_process_name
    cgroup_root = cgroups_root

    # Find out which type it is
    if process_name == 'hdfs-NAMENODE':
      return NameNodeMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hdfs-SECONDARYNAMENODE':
      return SecondaryNameNodeMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hdfs-DATANODE':
      return DataNodeMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'mapreduce-JOBTRACKER':
      return JobTrackerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'mapreduce-TASKTRACKER':
      return TaskTrackerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hbase-REGIONSERVER':
      return RegionServerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hbase-MASTER':
      return MasterMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'zookeeper-server':
      return ZooKeeperServerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'flume-AGENT':
      return FlumeAgentMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'cloudera-mgmt-SERVICEMONITOR':
      return ServiceMonitorMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'cloudera-mgmt-ACTIVITYMONITOR':
      return ActivityMonitorMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'cloudera-mgmt-EVENTSERVER':
      return EventServerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'cloudera-mgmt-ALERTPUBLISHER':
      return AlertPublisherMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'cloudera-mgmt-REPORTSMANAGER':
      return ReportsManagerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'cloudera-mgmt-HOSTMONITOR':
      return HostMonitorMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'cloudera-mgmt-NAVIGATOR':
      return NavigatorMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'impala-IMPALAD':
      return ImpaladMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'impala-STATESTORE':
      return StateStoreMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hdfs-JOURNALNODE':
      return JournalNodeMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hbase-HBASETHRIFTSERVER':
      return HbaseThriftServerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hbase-HBASERESTSERVER':
      return HbaseRestServerMonitor(self, self._registry, cgroup_root, cgroup_name)
    elif process_name == 'hdfs-HTTPFS':
      return HttpFSMonitor(self, self._registry, cgroup_root, cgroup_name)
    # This is really weird, but it works for now -- we use the same monitoring
    # code for both the MR and the HDFS failover controllers. We can get away
    # with this because we have no MR nor HDFS specific FC monitoring code. We
    # need to do this to avoid OPSAPS-13585.
    elif (process_name == 'hdfs-FAILOVERCONTROLLER' or
          process_name == 'mapreduce-FAILOVERCONTROLLER'):
      return FailoverControllerMonitor(self, self._registry, cgroup_root, cgroup_name)

    return None

  def get_registry_for_testing(self):
    return self._registry

  def get_metrics_url_timeout_seconds(self):
    return self._metrics_url_timeout_seconds

  def get_task_metrics_timeout_seconds(self):
    return self._task_metrics_timeout_seconds

  def _schedule(self):
    # Check if we need to schedule a full report, and determine the starting
    # point for our sleep time.
    sleep_time = self._time_until_next_full_report()
    if sleep_time <= 0.0:
      self._full_report = True
      sleep_time = self.get_reporting_interval_seconds()

    # If a new monitor has just become ready, we may need to send a partial
    # report (this will be ignored if we're already going to send a full
    # report). At the same time, we check the number of new monitors that
    # we will have after such a report is sent. If there are any, then we will
    # shorten our sleep, so we can quickly detect if those monitors become
    # ready. This call also clears any expired monitors from the map of
    # new monitors.
    num_monitors_ready, num_new_monitors = self._update_new_monitors()
    if num_new_monitors > 0:
      sleep_time = 1.0

    if self._full_report:
      self._reporting_thread.wake(CountDownLatch(1))
      self._last_full_report_scheduled = time.time()
    elif num_monitors_ready > 0:
      self._reporting_thread.wake(CountDownLatch(1))
      
    self._scheduling_thread.set_timeout(sleep_time)

  # This computes the time until the next full report should be sent. If a full
  # report needs to be sent now, then this will return a negative number.
  #
  # This is called by the scheduling thread.
  def _time_until_next_full_report(self):
    if self._last_full_report_scheduled is None:
      return -1
    return self._last_full_report_scheduled + \
           self.get_reporting_interval_seconds() - \
           time.time()

  # This call checks in on any new monitors that we have to see if they are
  # ready to report. It also checks if any new monitors are expired. This
  # returns a tuple of the number of monitors that are ready and the number
  # of new monitors that remain after the mutations this call has made.
  #
  # This is called by the scheduling thread.
  def _update_new_monitors(self):
    self._new_monitors_lock.acquire()
 
    monitors_to_remove = []
    for monitor, expiration in self._new_monitors.iteritems():
      if monitor.is_ready_to_report():
        LOG.info("Monitor ready to report: %s", (monitor,))
        monitors_to_remove.append(monitor)
        self._new_monitors_ready.append(monitor)
      elif time.time() > expiration:
        LOG.info("Monitor expired: %s", (monitor,))
        monitors_to_remove.append(monitor)

    for monitor in monitors_to_remove:
      del self._new_monitors[monitor]

    monitors_ready = len(self._new_monitors_ready)
    monitors_remaining = len(self._new_monitors)

    self._new_monitors_lock.release()

    return monitors_ready, monitors_remaining

  # Called to indicate that a new monitor has been created. This stores the new
  # monitor in the new_monitors map and wakes the scheduling thread.
  def new_monitor(self, monitor):
    LOG.info("New monitor: %s", (monitor,))
    self._new_monitors_lock.acquire()
    self._new_monitors[monitor] = time.time() + 60.0
    self._new_monitors_lock.release()
    self._scheduling_thread.wake(CountDownLatch(1))
